import re
import cairo
import codecs
import time
import os
import html5lib
import xml.dom.minidom
from xml.dom.minidom import Node

#import yaml
import syck as yaml # lots faster

def escapeJS(str):
	str = str.replace('\\', '\\\\').replace('"', '\\"')
	str = re.sub(r'\[(\w+)\]', r'[\\""+(\1)+"\\"]', str) # kind of an ugly hack, for nicer failure-message output
	return str

def escapeHTML(str):
	return str.replace('&', '&amp;').replace('<', '&lt;').replace('>', '&gt;').replace('"', '&quot;')

templates = yaml.load(open('templates.yaml').read())

spec_assertions = []
for s in yaml.load(open('spec.yaml').read())['assertions']:
	if 'meta' in s:
		eval(compile(s['meta'], '<meta spec assertion>', 'exec'), {}, {'assertions':spec_assertions})
	else:
		spec_assertions.append(s)

tests = []
for t in sum([ yaml.load(open(f).read()) for f in ['tests.yaml', 'tests2d.yaml', 'testsmisc.yaml']], []):
	if 'DISABLED' in t:
		continue
	if 'meta' in t:
		eval(compile(t['meta'], '<meta test>', 'exec'), {}, {'tests':tests})
	else:
		tests.append(t)

category_names = []
category_contents_direct = {}
category_contents_all = {}

spec_ids = {}
for t in spec_assertions: spec_ids[t['id']] = True
spec_refs = {}

def backref_html(name):
	backrefs = []
	c = ''
	for p in name.split('.')[:-1]:
		c += '.'+p
		backrefs.append('<a href="index%s.html">%s</a>.' % (c, p))
	backrefs.append(name.split('.')[-1])
	return ''.join(backrefs)

def make_flat_image(filename, w, h, r,g,b,a):
	if os.path.exists('tests/%s' % filename):
		return filename
	surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, w, h)
	cr = cairo.Context(surface)
	cr.set_source_rgba(r, g, b, a)
	cr.rectangle(0, 0, w, h)
	cr.fill()
	surface.write_to_png('tests/%s' % filename)
	return filename
	

try: os.mkdir('tests')
except: pass # ignore if it already exists

for i in range(len(tests)):
	test = tests[i]

	name = test['name']
	print "\r(%s)" % name, " "*32, "\t",
	
	cat_total = ''
	for cat_part in [''] + name.split('.')[:-1]:
		cat_total += cat_part+'.'
		if not cat_total in category_names: category_names.append(cat_total)
		category_contents_all.setdefault(cat_total, []).append(name)
	category_contents_direct.setdefault(cat_total, []).append(name)

	for ref in test.get('testing', []):
		if ref not in spec_ids:
			print "Test %s uses nonexistent spec point %s" % (name, ref)
		spec_refs.setdefault(ref, []).append(name)
	if not (len(test.get('testing', [])) or 'mozilla' in test):
		print "Test %s doesn't refer to any spec points" % name
	
	code = test['code']
	
	code = re.sub(r'@assert pixel (\d+,\d+) == (\d+,\d+,\d+,\d+);',
			r'_assertPixel(canvas, \1, \2, "\1", "\2");',
			code)

	code = re.sub(r'@assert pixel (\d+,\d+) ==~ (\d+,\d+,\d+,\d+);',
			r'_assertPixelApprox(canvas, \1, \2, "\1", "\2", 2);',
			code)

	code = re.sub(r'@assert pixel (\d+,\d+) ==~ (\d+,\d+,\d+,\d+) \+/- (\d+);',
			r'_assertPixelApprox(canvas, \1, \2, "\1", "\2", \3);',
			code)

	code = re.sub(r'@assert throws (\S+_ERR) (.*);',
			lambda m: 'try { var _thrown = false;\n  %s;\n} catch (e) { if (e.code != DOMException.%s) _fail("Failed assertion: expected exception of type %s, got: "+e.message); _thrown = true; } finally { _assert(_thrown, "should throw exception of type %s: %s"); }'
				% (m.group(2), m.group(1), m.group(1), m.group(1), escapeJS(m.group(2)))
			, code)

	code = re.sub(r'@assert throws (.*);',
			lambda m: 'try { var _thrown = false; %s; } catch (e) { _thrown = true; } finally { _assert(_thrown, "should throw exception: %s"); }'
				% (m.group(1), escapeJS(m.group(1)))
			, code)

	code = re.sub(r'@assert (.*) === (.*);',
			lambda m: '_assertSame(%s, %s, "%s", "%s");'
				% (m.group(1), m.group(2), escapeJS(m.group(1)), escapeJS(m.group(2)))
			, code)
	
	code = re.sub(r'@assert (.*) !== (.*);',
			lambda m: '_assertDifferent(%s, %s, "%s", "%s");'
				% (m.group(1), m.group(2), escapeJS(m.group(1)), escapeJS(m.group(2)))
			, code)
	
	code = re.sub(r'@assert (.*) == (.*);',
			lambda m: '_assertEqual(%s, %s, "%s", "%s");'
				% (m.group(1), m.group(2), escapeJS(m.group(1)), escapeJS(m.group(2)))
			, code)
	
	code = re.sub(r'@assert (.*) =~ (.*);',
			lambda m: '_assertMatch(%s, %s, "%s", "%s");'
				% (m.group(1), m.group(2), escapeJS(m.group(1)), escapeJS(m.group(2)))
			, code)

	code = re.sub(r'@assert (.*);',
			lambda m: '_assert(%s, "%s");'
				% (m.group(1), escapeJS(m.group(1)))
			, code)
	
	code = re.sub(r'@manual;', '_requireManualCheck();', code)
	#code = re.sub(r'@compare;', '_compareToExpected(canvas, document.getElementById("expected"));', code)

	assert(not re.match('@', code));

	background = ''
	if 'background' in test:
		background = ' style="background: %s"' % test['background']

	expectation_html = ''
	if 'expected' in test:
		expected = test['expected']
		expected_img = None
		if expected == 'none':
			pass
		elif expected == 'green':
			expected_img = make_flat_image('green-100x50.png', 100, 50, 0,1,0,1)
		else:
			if ';' in expected: print "Found semicolon in %s" % name
			expected = re.sub(r'^size (\d+) (\d+)',
				r'surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, \1, \2)\ncr = cairo.Context(surface)',
				expected)
			expected += "\nsurface.write_to_png('tests/%s.png')\n" % name
			eval(compile(expected, '<test %s>' % test['name'], 'exec'), {}, {'cairo':cairo})
			expected_img = "%s.png" % name

		if expected_img:
			expectation_html = ('<p class="output expectedtext">Expected output:' +
				'<p><img src="%s" class="output expected" id="expected" alt=""%s>' % (expected_img, background))

	canvas = test.get('canvas', 'width="100" height="50"') + background

	prev = tests[i-1]['name'] if i != 0 else 'index'
	next = tests[i+1]['name'] if i != len(tests)-1 else 'index'

	name_wrapped = name.replace('.', '.&#8203;') # (see https://bugzilla.mozilla.org/show_bug.cgi?id=376188)

	refs = ''.join('<li><a href="spec.html#testrefs.%s">%s</a>\n' % (n,n) for n in test.get('testing', []))
	if 'mozilla' in test:
		refs += '<li><a href="https://bugzilla.mozilla.org/show_bug.cgi?id=%d">Bugzilla</a>' % test['mozilla']

	notes = '<p class="notes">%s' % test['notes'] if 'notes' in test else ''

	images = ''
	for i in test.get('images', []):
		id = i.split('/')[-1]
		if '/' not in i:
			i = '../images/%s' % i
		images += '<img src="%s" id="%s" class="resource">\n' % (i,id)

	fallback = test.get('fallback', '<p class="fallback">FAIL (fallback content)</p>')

	template_params = {
		'name':name, 'name_wrapped':name_wrapped, 'backrefs':backref_html(name),
		'prev':prev, 'next':next, 'refs':refs, 'notes':notes, 'images':images,
		'canvas':canvas, 'expected':expectation_html, 'code':code, 'fallback':fallback
	}

	f = codecs.open('tests/%s.html' % name, 'w', 'utf-8')
	f.write(templates['standalone'] % template_params)
	
	f = codecs.open('tests/framed.%s.html' % name, 'w', 'utf-8')
	f.write(templates['framed'] % template_params)

	f = codecs.open('tests/minimal.%s.html' % name, 'w', 'utf-8')
	f.write(templates['minimal'] % template_params)


def write_index():
	f = open('tests/index.html', 'w')
	f.write(templates['index'] % { 'updated':time.strftime('%Y-%m-%d', time.gmtime()) })
	f.write('\n<!-- WebKit bugs break the "expand" feature -->\n<ul class="testlist">\n')
	depth = 1
	for category in category_names:
		name = category[1:-1] or ''
		count = len(category_contents_all[category])
		new_depth = category.count('.')
		while new_depth < depth: f.write(' '*(depth-1) + '</ul>\n'); depth -= 1
		f.write(' '*depth + templates['index.category.item'] % (name or 'all', name, count, '' if count==1 else 's'))
		while new_depth+1 > depth: f.write(' '*depth + '<ul>\n'); depth += 1
		for item in category_contents_direct.get(category, []):
			f.write(' '*depth + '<li><a href="%s.html">%s</a>\n' % (item, item) )
	while 0 < depth: f.write(' '*(depth-1) + '</ul>\n'); depth -= 1

def write_category_indexes():
	for category in category_names:
		name = (category[1:-1] or 'all')
		
		f = open('tests/index.%s.html' % name, 'w')
		f.write(templates['index.frame'] % { 'backrefs':backref_html(name), 'category':name })
		for item in category_contents_all[category]:
			f.write(templates['index.frame.item'] % item)

def write_reportgen():
	f = open('tests/reportgen.html', 'w')
	items_text = ''
	for item in category_contents_all['.']:
		items_text += templates['reportgen.item'] % { 'item':item }
	f.write(templates['reportgen'] % {'items':items_text })

def write_results():
	results = {}
	uas = []
	for item in category_contents_all['.']: results[item] = {}

	f = open('tests/results.html', 'w')
	f.write(templates['results'])

	for resultset in yaml.load(open('results2.yaml').read()):
		title = "%s (%s)" % (resultset['ua'], resultset['time'])
		assert title not in uas # don't allow repetitions
		uas.append(title)
		for r in resultset['results']:
			if r['id'] not in results:
				print 'Skipping results for removed test %s' % r['id']
				continue
			results[r['id']][title] = (
					r['status'].lower(),
					re.sub(r'%(..)', lambda m: chr(int(m.group(1), 16)), r['notes']))

	for ua in uas:
		f.write('<th>%s\n' % ua)
	for id in category_contents_all['.']:
		f.write('<tr><td><a href="%s.html">%s</a>\n' % (id, id))
		for ua in uas:
			status, details = results[id].get(ua, ('', ''))
			f.write('<td class="r %s"><ul class="d">%s</ul>\n' % (status, details))
	f.write('</table>\n')

def getNodeText(node):
	t, offsets = '', []

	# Skip over any previous annotations we added
	if node.nodeType == node.ELEMENT_NODE and 'testrefs' in node.getAttribute('class').split(' '):
		return t, offsets

	if node.nodeType == node.TEXT_NODE:
		val = node.nodeValue
		val = val.replace(unichr(0xa0), ' ') # replace &nbsp;s
		t += val
		offsets += [ (node, len(node.nodeValue)) ]
	for n in node.childNodes:
		child_t, child_offsets = getNodeText(n)
		t += child_t
		offsets += child_offsets
	return t, offsets

def html5Serializer(element):
	element.normalize()
	rv = []
	specialtext = ['style', 'script', 'xmp', 'iframe', 'noembed', 'noframes', 'noscript']
	empty = ['area', 'base', 'basefont', 'bgsound', 'br', 'col', 'embed', 'frame',
			 'hr', 'img', 'input', 'link', 'meta', 'param', 'spacer', 'wbr']

	def serializeElement(element):
		if element.nodeType == Node.DOCUMENT_TYPE_NODE:
			rv.append("<!DOCTYPE %s>" % element.name)
		elif element.nodeType == Node.DOCUMENT_NODE:
			for child in element.childNodes:
				serializeElement(child)
		elif element.nodeType == Node.COMMENT_NODE:
			rv.append("<!--%s-->" % element.nodeValue)
		elif element.nodeType == Node.TEXT_NODE:
			unescaped = False
			n = element.parentNode
			while n is not None:
				if n.nodeName in specialtext:
					unescaped = True
					break
				n = n.parentNode
			if unescaped:
				rv.append(element.nodeValue)
			else:
				rv.append(escapeHTML(element.nodeValue))
		else:
			rv.append("<%s" % element.nodeName)
			if element.hasAttributes():
				for name, value in element.attributes.items():
					rv.append(' %s="%s"' % (name, escapeHTML(value)))
			rv.append(">")
			if element.nodeName not in empty:
				for child in element.childNodes:
					serializeElement(child)
				rv.append("</%s>" % element.nodeName)
	serializeElement(element)
	return '<!DOCTYPE HTML>\n' + ''.join(rv)

def write_annotated_spec():
	# Load the stripped-down XHTMLised copy of the spec
	doc = xml.dom.minidom.parse(open('current-work-canvas.xhtml', 'r'))

	# Insert our new stylesheet
	n = doc.getElementsByTagName('head')[0].appendChild(doc.createElement('link'))
	n.setAttribute('rel', 'stylesheet')
	n.setAttribute('href', '../spectest.css')
	n.setAttribute('type', 'text/css')

	spec_assertion_patterns = []
	for a in spec_assertions:
		# Warn about problems
		if a['id'] not in spec_refs:
			print "Unused spec statement %s" % a['id']

		# Extract the marked keywords, and remove the markers
		pattern_text = a['text']
		keyword = 'none'
		for kw in ['must', 'should', 'required']:
			if ('*%s*' % kw) in pattern_text:
				keyword = kw
				pattern_text = pattern_text.replace('*%s*' % kw, kw)
				break

		# Convert the special pattern format into regexp syntax
		pattern_text = (pattern_text.
			# Escape relevant characters
			replace('*', r'\*').
			replace('+', r'\+').
			replace('.', r'\.').
			replace('(', r'\(').
			replace(')', r'\)').
			replace('[', r'\[').
			replace(']', r'\]').
			# Convert special sequences back into unescaped regexp code
			replace(' ', r'\s+').
			replace(r'<\.\.\.>', '.+').
			replace('<^>', '()').
			replace('<eol>', '\s*?\n')
		)
		pattern = re.compile(pattern_text, re.S)
		spec_assertion_patterns.append( (a['id'], pattern, keyword, a.get('previously', None)) )
	matched_assertions = {}

	for e in doc.getElementsByTagName('body')[0].childNodes:
		t, offsets = getNodeText(e)
		for id, pattern, keyword, previously in spec_assertion_patterns:
			m = pattern.search(t)
			if m:
				# When the pattern-match isn't enough to uniquely identify a sentence,
				# allow explicit back-references to earlier paragraphs
				if previously:
					n, text = previously
					node = e
					while n:
						node = node.previousSibling
						n -= 1
					if text not in getNodeText(node)[0]:
						continue # discard this match

				if id in matched_assertions:
					print "Spec statement %s matches multiple places" % id
				matched_assertions[id] = True

				end = m.end(1)
				end_node = None
				for end_node, o in offsets:
					if end < o:
						break
					end -= o
				assert(end_node)

				n1 = doc.createElement('span')
				n1.setAttribute('class', 'testrefs kw-%s' % keyword)
				n1.setAttribute('id', 'testrefs.%s' % id)
				n1.appendChild(doc.createTextNode(' '))
				
				n = n1.appendChild(doc.createElement('a'))
				n.setAttribute('href', '#testrefs.%s' % id)
				n.setAttribute('title', id)
				n.appendChild(doc.createTextNode('#'))

				n1.appendChild(doc.createTextNode(' '))
				for test_id in spec_refs.get(id, []):
					n = n1.appendChild(doc.createElement('a'))
					n.setAttribute('href', '%s.html' % test_id)
					n.appendChild(doc.createTextNode(test_id))
					n1.appendChild(doc.createTextNode(' '))
				n0 = doc.createTextNode(end_node.nodeValue[:end])
				n2 = doc.createTextNode(end_node.nodeValue[end:])

				p = end_node.parentNode
				p.replaceChild(n2, end_node)
				p.insertBefore(n1, n2)
				p.insertBefore(n0, n1)

				t, offsets = getNodeText(e)

	for s in spec_assertions:
		if s['id'] not in matched_assertions:
			print "Annotation incomplete: Unmatched spec statement %s" % s['id']

	# Convert from XHTML5 back to HTML5
	doc.documentElement.removeAttribute('xmlns')
	doc.documentElement.setAttribute('lang', doc.documentElement.getAttribute('xml:lang'))

	head = doc.documentElement.getElementsByTagName('head')[0]
	head.insertBefore(doc.createElement('meta'), head.firstChild).setAttribute('charset', 'UTF-8')

	codecs.open('tests/spec.html', 'w', 'utf-8').write(html5Serializer(doc))


write_index()
write_category_indexes()
write_reportgen()
write_results()
write_annotated_spec()

